# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import cast
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from .resourcestate import CurrentResourceState
from .resourcestate import ResourceState
from .versionedresource import VersionedResource


C = TypeVar('C', bound=ResourceState[Any] | CurrentResourceState)
S = TypeVar('S', bound=pydantic.BaseModel)


class TransitioningResource(VersionedResource[S], Generic[S, C]):
    status: C = pydantic.Field(
        default=None,
        description=(
            "Describes the current state of the resource. "
        ),
    )

    def _initialize_status(self) -> None:
        status = self.initialize_status()
        if status is None:
            field = self.model_fields['status']
            adapter = cast(
                pydantic.TypeAdapter[C],
                pydantic.TypeAdapter(field.annotation)
            )
            status = adapter.validate_python({})
        self.status = status

    def initialize_status(self) -> C | None:
        return None

    def model_post_init(self, _: Any) -> None:
        if self.status is None: # type: ignore
            self._initialize_status()

    def set_status(
        self,
        status: str,
        *,
        timestamp: datetime.datetime | None = None,
        message: str = '',
        **params: Any
    ) -> C:
        self.logger.debug(
            "%s status changed (pk: %s, status: %s, message: %s)",
            type(self).__name__,
            self.pk,
            status,
            message
        )
        return self.status.new(
            status=status,
            generation=self.metadata.generation,
            message=message,
            timestamp=timestamp or datetime.datetime.now(datetime.timezone.utc),
            **{**self.new_status_params(), **params}
        )

    def new_status_params(self) -> dict[str, Any]:
        return {}


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Callable
from typing import TypeVar


T = TypeVar('T')


class IMaskable(str):
    __module__: str = 'canonical.protocols'

    def mask(self, masker: Callable[[bytes], T]) -> T:
        return masker(str.encode(f'phonenumber:{self.lower()}', 'ascii'))


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Generator

import iso3166
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema


# Corrections for wrong names
COUNTRY_NAMES: dict[str, str] = {
    'NL': 'The Netherlands'
}

# Additional codes that are not in the iso3166 library (TODO: make issue at Github)
_records: list[iso3166.Country] = [
    # This one is problematic because it only has an ISO 3166 Alpha 2 code.
    iso3166.Country("Northern Ireland", "XI", 'GBR', '826', "Northern Ireland"),
]

# Another EU peculiarity: The European Commission generally uses
# ISO 3166-1 alpha-2 codes with two exceptions: EL (not GR) is
# used to represent Greece, and UK (not GB) is used to represent
# the United Kingdom.[10][11] This notwithstanding, the Official
# Journal of the European Communities specified that GR and GB be
# used to represent Greece and United Kingdom respectively.[12]
# For VAT administration purposes, the European Commission uses
# EL and GB for Greece and the United Kingdom respectively.
# https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2.
_eu_alpha2_mapping: dict[str, iso3166.Country] = {
    'EL': iso3166.countries_by_alpha2['GR'],
    'UK': iso3166.countries_by_alpha2['GB']
}

_alpha2_eu_mapping: dict[str, str] = {v.alpha2: k for k, v in _eu_alpha2_mapping.items()}

countries_by_alpha2: dict[str, iso3166.Country] = {
    x.alpha2: x
    for x in _records
}


class ISO3166Base(str):
    length: int

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., str | None], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, v: str, _: Any = None):
        if len(v) != cls.length:
            raise ValueError('Value is not an ISO 3166 code.')
        try:
            return cls(cls._get_country(v))
        except LookupError:
            raise ValueError(f"Invalid code: {v}")

    @classmethod
    def _get_country(cls, v: str) -> str:
        raise NotImplementedError


class ISO3166Alpha2(ISO3166Base):
    length: int = 2

    @property
    def name(self) -> str:
        if str(self) in COUNTRY_NAMES:
            return COUNTRY_NAMES[self]
        c = iso3166.countries_by_alpha2.get(self)
        if c is None: # pragma: no cover
            raise LookupError
        return c.name

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema(max_length=cls.length))

    @classmethod
    def _get_country(cls, v: str) -> str:
        c = countries_by_alpha2.get(v)\
            or iso3166.countries_by_alpha2.get(v)\
            or _eu_alpha2_mapping.get(v)
        if c is None:
            raise LookupError
        return c.alpha2


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import cast
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from .condition import Condition


C = TypeVar('C', bound=Condition[Any])



class CurrentResourceState(pydantic.BaseModel):
    """Describes the current state of a :class:`VersionedResource`."""

    changed: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        description=(
            "The date and time of the last change to this resource."
        )
    )

    version: int = pydantic.Field(
        default=0
    )

    def new(
        self,
        *,
        status: str,
        generation: int,
        timestamp: datetime.datetime,
        message: str = '',
        **params: Any
    ) -> Any:
        raise NotImplementedError


class ResourceState(CurrentResourceState, Generic[C]):
    """Like :class:`CurrentResourceState`, but maintains the historical
    states.
    """
    model_config = {'populate_by_name': True}

    dirty: bool = pydantic.Field(
        default=False,
        exclude=True
    )

    conditions: list[C] = pydantic.Field(
        default_factory=list,
        description=(
            "Describes the state of the resource in reverse "
            "chronological order (the first item in the array is "
            "the first known state)."
        )
    )

    message: str | None = pydantic.Field(
        default=None,
        description=(
            "The message of the last known condition."
        )
    )

    current: str | None = pydantic.Field(
        default=None,
        description=(
            "The status of the last known condition."
        )
    )

    def _apply(self, condition: C, replay: bool = False) -> C:
        if self.is_final():
            raise ValueError("Resource is in its final state.")
        if condition.timestamp == self.changed:
            condition.timestamp += datetime.timedelta(microseconds=1)
        self.changed = condition.timestamp
        self.message = condition.message
        self.current = condition.status
        self.apply(condition)
        if not replay:
            self.conditions.append(condition)
        if not replay:
            self.dirty = True
        return self.conditions[-1]

    def apply(self, condition: C) -> None:
        raise NotImplementedError

    def get_persistable_fields(self) -> set[str]:
        return set(ResourceState.model_fields.keys())

    def has(self, status: str) -> bool:
        return any([x.status == status for x in self.conditions])

    def model_post_init(self, _: Any) -> None:
        self.conditions = list(sorted(list(self.conditions), key=lambda x: x.timestamp.timestamp()))
        for condition in self.conditions:
            self._apply(condition, replay=True)

    def is_dirty(self) -> bool:
        return self.dirty

    def is_final(self) -> bool:
        return bool(self.conditions) and self.conditions[0].is_final()

    def new(
        self,
        *,
        status: str,
        generation: int,
        timestamp: datetime.datetime,
        message: str = '',
        **params: Any
    ) -> C:
        field = self.model_fields['conditions']
        adapter = cast(
            pydantic.TypeAdapter[list[C]],
            pydantic.TypeAdapter(field.annotation)
        )
        condition, *extra = adapter.validate_python([{
            **self.new_status_params(),
            **params,
            'message': message,
            'observed_generation': generation,
            'status': status,
            'timestamp': timestamp
        }])
        if extra:
            raise ValueError(f"{type(self).__name__}.new() creates one condition.")
        return self.update(condition)

    def new_status_params(self) -> dict[str, Any]:
        return {}

    def update(self, condition: C, replay: bool = False) -> C:
        return self._apply(condition, replay=replay)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import Mapping
from typing import TypeVar

import pydantic

from .protocols import IRepository
from .objectmeta import ObjectMeta


__all__: list[str] = [
    'VersionedResource'
]

L = TypeVar('L')
S = TypeVar('S', bound=pydantic.BaseModel)
T = TypeVar('T', bound='VersionedResource[Any]')
V = TypeVar('V')
NOT_PROVIDED: object = object()


class VersionedResource(pydantic.BaseModel, Generic[S]):
    model_config = {'populate_by_name': True}
    logger: ClassVar[logging.Logger] = logging.getLogger('canonical.resource')
    version: ClassVar[str] = 'v1'
    _storage: IRepository['VersionedResource[S]'] = pydantic.PrivateAttr()

    api_version: str = pydantic.Field(
        default=...,
        alias='apiVersion',
        title="API Version",
        description=(
            "The `apiVersion` field defines the versioned schema of this "
            "representation of an object. Servers should convert recognized "
            "schemas to the latest internal value, and may reject "
            "unrecognized values."
        )
    )

    kind: str = pydantic.Field(
        default=...,
        description=(
            "Kind is a string value representing the REST resource this "
            "object represents. Servers may infer this from the endpoint "
            "the client submits requests to. Cannot be updated. In CamelCase."
        )
    )

    metadata: ObjectMeta = pydantic.Field(
        default=...,
        description=(
            "`ObjectMeta` is metadata that all persisted resources "
            "must have, which includes all objects users must create."
        )
    )

    spec: S = pydantic.Field(
        default=...,
        description="Specification of the resource."
    )

    @property
    def pk(self) -> Any:
        return self.get_primary_key()

    @classmethod
    def new(
        cls: type[T],
        name: str,
        spec: Mapping[str, Any],
        labels: Mapping[str, Any] | None = None,
        **kwargs: Any
    ) -> T:
        metadata: Mapping[str, Any] = {
            'name': name,
            'labels': labels or {}
        }
        params: Mapping[str, Any] = {
            'api_version': cls.version,
            'kind': cls.__name__,
            'metadata': metadata,
            'spec': spec
        }
        self = cls.model_validate({**kwargs, **params})
        cls.logger.debug("Initialized new %s/%s (pk: %s)", self.api_version, cls.__name__, self.pk)
        return self

    def attach(self, repo: IRepository[Any]):
        self._storage = repo
        return self

    def clone(self, **spec: Any):
        cloned = self.model_validate({
            **self.model_dump(),
            'spec': {**self.spec.model_dump(), **spec}
        })
        cloned.set_label('cloned', 'true')
        return cloned

    def get_label(self, name: str, factory: Callable[[Any], L] = str) -> L | None:
        adapter = pydantic.TypeAdapter(factory)
        return adapter.validate_python(self.metadata.labels.get(name)) # type: ignore

    def get_primary_key(self) -> Any:
        raise NotImplementedError

    def is_cloned(self) -> bool:
        l = self.get_label('cloned')
        return isinstance(l, bool) or l == 'true'

    def is_labeled(self, name: str, *names: str) -> bool:
        return self.metadata.is_labeled([name, *names])

    def set_label(self, name: str, value: Any) -> Any | None:
        return self.metadata.set_label(name, value)

    async def persist(self):
        await self._storage.persist(self)
        return self


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from .addressee import Addressee
from .baseaddress import BaseAddress


class Address(BaseAddress):
    model_config = {'populate_by_name': True}

    kind: Literal['Address'] = pydantic.Field(
        default='Address'
    )

    addressee: Addressee | None = pydantic.Field(
        default=None
    )

    remarks: str | None = pydantic.Field(
        default=None
    )



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .text import Text


class PersonalName(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    given_name: Text | None = pydantic.Field(
        default=None,
        alias='givenName',
        description=(
            "Given name(s) or first name(s). Note that in some cultures, "
            "people can have multiple given names; all can be present, "
            "with the names being separated by space characters."
        )
    )

    family_name: Text = pydantic.Field(
        default=...,
        alias='familyName',
        description=(
            "Surname(s) or last name(s). Note that in some cultures, people "
            "can have multiple family names or no family name; all can be "
            "present, with the names being separated by space characters.\n\n"
            "If there is no surname, then this field is considered the name and "
            "the `givenName` field must be empty."
        )
    )

    @property
    def display_name(self):
        if not self.given_name:
            return self.family_name
        return f'{self.given_name} {self.family_name}'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


__all__: list[str] = [
    'UnixTimestamp'
]


class UnixTimestamp(datetime.datetime):

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.int_schema(),
            python_schema=core_schema.chain_schema([
                core_schema.union_schema([
                    core_schema.is_instance_schema(datetime.datetime),
                    core_schema.chain_schema([
                        core_schema.union_schema([
                            core_schema.int_schema(),
                            core_schema.chain_schema([
                                core_schema.str_schema(),
                                core_schema.no_info_plain_validator_function(lambda x: int(x))
                            ])
                        ]),
                        core_schema.no_info_plain_validator_function(
                            lambda timestamp: datetime.datetime.fromtimestamp(
                                timestamp,
                                tz=datetime.timezone.utc
                            )
                        )
                    ])
                ]),
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(cls.serialize),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.int_schema())

    @classmethod
    def serialize(cls, value: int | datetime.datetime) -> int:
        if isinstance(value, datetime.datetime):
            value = int(value.timestamp())
        return value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum
from typing import Literal

import pydantic


__all__: list[str] = [
    'Incoterm',
    'IncotermEnum',
    'IncotermLiteral'
]


class Incoterm(pydantic.BaseModel):
    version: int = pydantic.Field(
        default=2020
    )

    term: 'IncotermLiteral' = pydantic.Field(
        default=...
    )


class IncotermEnum(str, enum.Enum):
    EXW = 'EXW'
    FCA = 'FCA'
    FAS = 'FAS'
    FOB = 'FOB'
    CPT = 'CPT'
    CIP = 'CIP'
    CFR = 'CFR'
    CIF = 'CIF'
    DAP = 'DAP'
    DPU = 'DPU'
    DDP = 'DDP'


IncotermLiteral = Literal[
    'EXW',
    'FCA',
    'FAS',
    'FOB',
    'CPT',
    'CIP',
    'CFR',
    'CIF',
    'DAP',
    'DPU',
    'DDP',
]


Incoterm.model_rebuild()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union

import pydantic

from .emailaddress import EmailAddress
from .honorific import HonorificEnum
from .phonenumber import Phonenumber
from .text import Text


AddresseeType = Union['PersonalAddressee', 'OrganizationAddressee']


class Addressee(pydantic.RootModel[AddresseeType]):

    @property
    def attention_name(self) -> str | None:
        return self.root.attention_name

    @property
    def display_name(self) -> str:
        return self.root.display_name


class OrganizationAddressee(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    organization_name: Text = pydantic.Field(
        default=...,
        alias='organizationName'
    )

    department: Text | None = pydantic.Field(
        default=None
    )

    attention: Union['PersonalAddressee', None] = pydantic.Field(
        default=None
    )

    email: EmailAddress | None = pydantic.Field(
        default=None
    )

    phonenumber: Phonenumber | str | None = pydantic.Field(
        default=None
    )

    @property
    def attention_name(self) -> str | None:
        return self.attention.display_name if self.attention else None

    @property
    def display_name(self) -> str:
        return self.organization_name


class PersonalAddressee(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    given_name: Text | None = pydantic.Field(
        default=None,
        alias='givenName'
    )

    family_name: Text = pydantic.Field(
        default=...,
        alias='familyName'
    )

    email: EmailAddress | None = pydantic.Field(
        default=None
    )

    phonenumber: Phonenumber | str | None = pydantic.Field(
        default=None
    )
    
    honorific: HonorificEnum | None = pydantic.Field(
        default=None
    )

    @property
    def attention_name(self) -> str | None:
        return None

    @property
    def display_name(self) -> str:
        if not self.given_name:
            return self.family_name
        return f'{self.given_name} {self.family_name}'


Addressee.model_rebuild()
OrganizationAddressee.model_rebuild()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from typing import cast
from typing import get_args
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Generic
from typing import TypeVar

from canonical import DisplayName
from canonical import EmailAddress
from canonical import ResourceName
from canonical import PersonalName
from canonical import Phonenumber


DefaultKeyType = ResourceName | EmailAddress | Phonenumber | int
C = TypeVar('C')
K = TypeVar('K')
R = TypeVar('R', bound=Any)


class BaseTransactionParty(pydantic.BaseModel, Generic[K, R]):
    """Describes a :term:`Party` that participates in
    a :term:`Transaction`.

    A :class:`TransactionParty` instance contains sufficient
    information to either retrieve or create a corresponding
    entity in an external system, such as a Customer Relationship
    Management (CRM) system.
    """
    model_config = {'populate_by_name': True}

    key: K = pydantic.Field(
        default=...,
        description=(
            "A data element that identifier the party. May be a "
            "`ResourceName`, email address, phone number or "
            "integer identifier."
        )
    )

    transaction: ResourceName | int | None = pydantic.Field(
        default=None,
        description=(
            "Optionaly identifies the transaction. This field may be "
            "left empty if the `TransactionParty` is part of a schema "
            "from which the transaction can be inferred through other "
            "properties."
        )
    )

    party_name: PersonalName | DisplayName = pydantic.Field(
        default=...,
        alias='partyName',
        description=(
            "The name of the party. For organizations, this field is a string;"
            " for natural persons this field is a `PersonalName "
            "object."
        )
    )

    roles: set[R] = pydantic.Field(
        default_factory=set,
        description=(
            "A set indicating the roles that the participating "
            "party has in the transaction."
        ),
        min_length=1
    )

    resource_links: set[ResourceName] = pydantic.Field(
        default_factory=set,
        alias='resourceLinks',
        description=(
            "An array of strings identifying the external resources linked to "
            "the participating party, such as a record in a CRM system."
        )
    )

    email: EmailAddress | None = pydantic.Field(
        default=None,
        description=(
            "The email address of the participating party, if known."
        )
    )

    phonenumber: Phonenumber | str | None = pydantic.Field(
        default=None,
        description=(
            "The phone number of the participating party, if known."
        )
    )

    @property
    def display_name(self) -> str:
        return self.party_name.display_name

    @property
    def role(self) -> R:
        if len(self.roles) > 1:
            raise ValueError("Participating party has multiple roles.")
        return list(self.roles)[0]

    def model_post_init(self, _: Any) -> None:
        if not self.email and isinstance(self.key, EmailAddress):
            self.email = self.key
        if not self.phonenumber and isinstance(self.key, Phonenumber):
            self.phonenumber = self.key
        if not self.roles:
            raise ValueError("At least one role must be specified.")

    def has_link(self, service: str, kind: str | None = None) -> bool:
        """Return a boolean indicating if the participating party is
        linked to an external resource in the given `service`, optionally
        further specified by the `kind` parameter.
        """
        for name in self.resource_links:
            if name.service != service:
                continue
            if kind is not None and name.kind != kind:
                continue
            result = True
            break
        else:
            result = False
        return result

    def has_role(self, role_name: str) -> bool:
        """Return a boolean indicating if the party participates in
        the transaction with the given role.
        """
        if not get_args(self.model_fields['roles'].annotation) == (str,):
            raise NotImplementedError("Subclasses must override this method.")
        return role_name in cast(set[str], self.roles)

    async def retrieve(
        self,
        func: Callable[..., Awaitable[C]]
    ) -> C:
        """Retrieve a record from an external system."""
        return await func(self)


class TransactionParty(BaseTransactionParty[DefaultKeyType, str]):
    model_config = {'populate_by_name': True}